/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "internal.h"

/* ********************************************************************** */
/* The UTF-8 decoder was written by Bjoern Hoehrmann.   Here is his license
 * for his decoder:
 *
 * Copyright (c) 2008-2009 Bjoern Hoehrmann <bjoern@hoehrmann.de>
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * And there is his code, slightly modified to compile in this environment:
 */
/* BEGIN UTF-8 DECODER */
/* Copyright (c) 2008-2009 Bjoern Hoehrmann <bjoern@hoehrmann.de>
 * See http://bjoern.hoehrmann.de/utf-8/decoder/dfa/ for details. */

static unsigned char const utf8d_a[256]= {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 00..1f */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 20..3f */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 40..5f */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 60..7f */
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9, /* 80..9f */
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7, /* a0..bf */
  8,8,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, /* c0..df */
  0xa,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x4,0x3,0x3, /* e0..ef */
  0xb,0x6,0x6,0x6,0x5,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8, /* f0..ff */
};
static unsigned char const utf8d_b[]= {
  0x0,0x1,0x2,0x3,0x5,0x8,0x7,0x1,0x1,0x1,0x4,0x6,0x1,0x1,0x1,0x1, /* s0..s0 */
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1, /* s1..s2 */
  1,2,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1, /* s3..s4 */
  1,2,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,3,1,1,1,1,1,1, /* s5..s6 */
  1,3,1,1,1,1,1,3,1,3,1,1,1,1,1,1,1,3,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* s7..s8 */
};

static INLINE LiqChar32 utf8_decode (LiqChar32 *state, LiqChar32 *codep, LiqChar32 byte)
{
    LiqChar32 type= utf8d_a[byte];

    *codep= (*state != UTF_ACCEPT) ?
        (byte & 0x3f) | (*codep << 6) :
        (0xff >> type) & (byte);

    *state= utf8d_b[*state * 16 + type];
    return *state;
}

/* END UTF-8 DECODER */
/* ********************************************************************** */

static size_t liq_utf8_strnlen (LiqChar8 const *s, size_t n)
{
    LiqChar32 cp;
    LiqChar32 state= UTF_ACCEPT;
    size_t    len= 0;
    while (n--) {
        if (utf8_decode(&state, &cp, (LiqUInt8)*s++) == UTF_ACCEPT)
            len++;
    }
    if (state != UTF_ACCEPT) {
        assert (0 && "bad utf-8 encoding"); /* die in debug mode */
        return 0;                           /* reject string completely in production code */
    }
    return len;
}

void liq_render_set_text_n_utf8 (LiqRender *self, LiqChar8 const *s, size_t n)
{
    liq_render_alloc_info (self, liq_utf8_strnlen (s, n));
    {
        LiqCharInfo *c= self->m_info;
        LiqChar32 cp= 0;
        LiqChar32 state= UTF_ACCEPT;
        while (n--) {
            if (utf8_decode(&state, &cp, (LiqUInt8)*s++) == UTF_ACCEPT)
                liq_set_info (c++, cp);
        }
    }
    liq_render_finalise_text (self);
}

void liq_render_set_text_utf8 (LiqRender *self, LiqChar8 const *text)
{
    liq_render_set_text_n_utf8 (self, text, strlen (text));
}

